/*
Copyright (c) 2005-2006, MentorGen, LLC
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

+ Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.
+ Redistributions in binary form must reproduce the above copyright notice, 
  this list of conditions and the following disclaimer in the documentation 
  and/or other materials provided with the distribution.
+ Neither the name of MentorGen LLC nor the names of its contributors may be 
  used to endorse or promote products derived from this software without 
  specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
  POSSIBILITY OF SUCH DAMAGE.
 */
package com.mentorgen.tools.profile.instrument;

import org.objectweb.asm.jip.Label;
import org.objectweb.asm.jip.MethodAdapter;
import org.objectweb.asm.jip.MethodVisitor;
import org.objectweb.asm.jip.Opcodes;

import com.mentorgen.tools.profile.Controller;

import static org.objectweb.asm.jip.Opcodes.INVOKESTATIC;

/**
 * This class is responsible for instrumenting a method to 
 * call the profiler in order for performance
 * data to be gathered. The basic idea is that the profiler is called
 * when a method starts and when it exists which allows the profiler
 * to gather performance data (note that a method can be exited from
 * when an exception is thrown as well as when return is called). The
 * one big caveate is static initializers. They are not called as part
 * of the flow of the program &mdash; they are called by the classloader.
 * Since they whole premise of the profiler is built on the idea of a
 * orderly call stack, static initializers are not instrumented. 
 * 
 * @author Andrew Wilcox
 *
 */
public class PerfMethodAdapter extends MethodAdapter {
	private String _className, _methodName;
	private boolean _clinit = false;
	private boolean _init = false;
	//EM 22/10/2008
	private String superName;
	private boolean isFinalize = false;
	private boolean superFinalizeCalled = false;

	public PerfMethodAdapter(MethodVisitor visitor, 
			String className,
			String methodName,
			//EM 22/10/2008
			String aSuperName
			) { 
		super(visitor);
		_className = className;
		_methodName = methodName;
		//EM 22/10/2008
		superName = aSuperName;

		// Static initializers are excluded. The reason for this
		// is the the profiling algorithm we're using mirrors the call stack.
		// Since static initializers are called by the classloader
		// and therefore aren't part of the programs flow of control,
		// static initializers can really mess up the profiler, especially
		// when they're called before the program's flow of control is started
		// (for example, the when the class with the main() method has a 
		// static initalizer). So yes, this is a short comming in the
		// design of the profiler, but we're willing to live with it because
		// this profiler is lightweight and allows us to use it interactively.
		//
		if (methodName.equals("<clinit>")) {
			_clinit = true;
		} else if (methodName.startsWith("<init>")) {
			_init = true;
		}
	}

	public void visitCode() {
		if (_clinit) {
			super.visitCode();
			return;
		} 
		
		//EM 28/10/2008 : directly use the method sign
		//this.visitLdcInsn(_className);
		//this.visitLdcInsn(_methodName);
		this.visitLdcInsn(_className + "." + _methodName);
		this.visitMethodInsn(INVOKESTATIC, 
				Controller._profiler, 
				"start", 
		//"(Ljava/lang/String;Ljava/lang/String;)V");
				"(Ljava/lang/String;)V");
		
		super.visitCode();
	}

	public void visitInsn(int inst) {
		if (_clinit) {
			super.visitInsn(inst);
			return;
		}

		switch (inst) {
		case Opcodes.ARETURN:
		case Opcodes.DRETURN:
		case Opcodes.FRETURN:
		case Opcodes.IRETURN:
		case Opcodes.LRETURN:
		case Opcodes.RETURN:
		case Opcodes.ATHROW:
			//EM 20/10/2008 : always track alloc, and after full init to get "this"
			if (/*Controller._trackObjectAlloc && */_init) {
				super.visitVarInsn(Opcodes.ALOAD, 0);
				super.visitLdcInsn(_className.replaceAll("/", "."));
				super.visitLdcInsn(_className);
				super.visitMethodInsn(INVOKESTATIC, 
						Controller._profiler, 
						"alloc", 
						"(Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V");
				
			}
			
			//EM 20/10/2008
			if(_methodName.equals("finalize")){
				isFinalize = true;
				super.visitVarInsn(Opcodes.ALOAD, 0);
				super.visitLdcInsn(_className.replaceAll("/", "."));
				super.visitMethodInsn(INVOKESTATIC, 
						Controller._profiler, 
						"gc", 
						"(Ljava/lang/Object;Ljava/lang/String;)V");			
			}
			
			//EM 22/10/2008
			if(isFinalize && !superFinalizeCalled){
//				System.out.println("ADD SUPER FINALIZE:" + _className + ":" + superName);
				//Add super finalize
//				Label l0 = new Label();
//				super.visitLabel(l0);
				super.visitVarInsn(Opcodes.ALOAD, 0);
				super.visitMethodInsn(Opcodes.INVOKESPECIAL, superName, "finalize", "()V");
			}
			
			//EM 28/10/2008 : Even, no need to have paramters here, only for debug
			//EM 28/10/2008 : directly use the method sign
			//this.visitLdcInsn(_className);
			//this.visitLdcInsn(_methodName);
			//this.visitLdcInsn(_className + "." + _methodName);
			this.visitMethodInsn(INVOKESTATIC, 
					Controller._profiler, 
					"end", 
					//"(Ljava/lang/String;Ljava/lang/String;)V");
			//"(Ljava/lang/String;)V");
			"()V");
			break;

		default:
			break;
		}
		
		if (Opcodes.MONITORENTER == inst) {
			this.visitLdcInsn(_className);
			this.visitLdcInsn(_methodName);
						
			this.visitMethodInsn(INVOKESTATIC, 
					Controller._profiler, 
					"beginWait", 
					"(Ljava/lang/String;Ljava/lang/String;)V");
			
			super.visitInsn(inst);

			this.visitLdcInsn(_className);
			this.visitLdcInsn(_methodName);
						
			this.visitMethodInsn(INVOKESTATIC, 
					Controller._profiler, 
					"endWait", 
					"(Ljava/lang/String;Ljava/lang/String;)V");			
		} else {
			super.visitInsn(inst);
		}
	}

	//EM 16/10/2008
	boolean newThreadViewed = false;
	
	@Override
	public void visitMethodInsn(int opcode, String owner, String name, String desc) {
		if (isWaitInsn(opcode, owner, name, desc)) {

			this.visitLdcInsn(_className);
			this.visitLdcInsn(_methodName);
						
			this.visitMethodInsn(INVOKESTATIC, 
					Controller._profiler, 
					"beginWait", 
					"(Ljava/lang/String;Ljava/lang/String;)V");
			
			super.visitMethodInsn(opcode, owner, name, desc);

			this.visitLdcInsn(_className);
			this.visitLdcInsn(_methodName);
						
			this.visitMethodInsn(INVOKESTATIC, 
					Controller._profiler, 
					"endWait", 
					"(Ljava/lang/String;Ljava/lang/String;)V");
		} else {
			//EM 22/10/2008 : detect call to super.finalize
			if(isFinalize && name.equals("finalize")){
				superFinalizeCalled = true;
			}
			//EM 16/10/2008
			if("java/lang/Thread".equals(owner) && name.equals("<init>")){// && newThreadViewed){
//				System.out.println("JOPed method : CN="+_className+" O="+owner+" N="+name+" D="+desc);
				super.visitMethodInsn(opcode, "em/jop/profiler/JOPThread", name, desc);
				newThreadViewed = false;
//				System.exit(0);
			}
//			else if(name.equals("<init>")){
//				System.out.println("METH: CN="+_className+" O="+owner+" N="+name+" D="+desc);
//				super.visitMethodInsn(opcode, owner, name, desc);
//				newThreadViewed = false;
//			}
			else{
				super.visitMethodInsn(opcode, owner, name, desc);
			}
		}
	}
	
	//
	// code to handle unwinding the call stack when an exception is thrown
	// (many thanks to Fredrik Svar�n for posting this code in the help forum!)
	//
	
	//EM 16/10/2008
	@Override
	public void visitTypeInsn(int opcode, String desc) {
		if(opcode==Opcodes.NEW && desc.equals("java/lang/Thread")){
			System.out.println("JOPed type new Thread() Thread -> JOPThread");
			newThreadViewed = true;
			super.visitTypeInsn(opcode, "em/jop/profiler/JOPThread");
		}
		else{
			super.visitTypeInsn(opcode, desc);
		}
	}

	@Override
	public void visitTryCatchBlock(Label start, Label end, Label handler, String type) {
		super.visitTryCatchBlock(start, end, handler, type);
		
		// Note: static initializers aren't measured, so make sure that the exception
		// isn't being caught in one
		if (type != null && !_clinit) {
			handler.info = new ExceptionInfo(type);
		}
	}
	
	@Override
	public void visitLabel(Label label) {
		super.visitLabel(label);
		
		if (label.info instanceof ExceptionInfo) {
			this.visitLdcInsn(_className);
			this.visitLdcInsn(_methodName);
			this.visitLdcInsn(((ExceptionInfo)label.info).type);

			this.visitMethodInsn(INVOKESTATIC, 
					Controller._profiler, 
					"unwind", 
					"(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V");
		}
	}

	class ExceptionInfo {
		String type;
		ExceptionInfo(String type) {
			this.type = type;
		}
	}
	
	//
	// private methods
	//

	private static boolean isWaitInsn(int opcode, String owner, String name, String desc) {
		boolean isWait = (opcode == Opcodes.INVOKEVIRTUAL 
				&& "java/lang/Object".equals(owner) 
				&& "wait".equals(name)
				&& ("()V".equals(desc) || "(J)V".equals(desc) || "(JI)V".equals(desc)));		
		if (isWait) return true;

		isWait = (opcode == Opcodes.INVOKEVIRTUAL
				&& "java/lang/Thread".equals(owner) 
				&& "join".equals(name)
				&& ("()V".equals(desc) || "(J)V".equals(desc) || "(JI)V".equals(desc)));
		if (isWait) return true;
		
		isWait = (opcode == Opcodes.INVOKESTATIC 
				&& "java/lang/Thread".equals(owner) 
				&& "sleep".equals(name)
				&& ("(J)V".equals(desc) || "(JI)V".equals(desc)));
		if (isWait) return true;
		
		isWait = (opcode == Opcodes.INVOKESTATIC 
				&& "java/lang/Thread".equals(owner) 
				&& "yield".equals(name)
				&& "()V".equals(desc));
		if (isWait) return true;
		
		return isWait;
	}

	//EM 20/10/2008
	@Override
	public void visitLocalVariable(String name, String desc, String signature,
			Label start, Label end, int index) {
		if(desc.equals("java/lang/Thread")){
			System.out.println("JOPed local variable "+name+" Thread -> JOPThread");
			super.visitLocalVariable(name, "em/jop/profiler/JOPThread", signature, start, end, index);
		}
		else{
		// TODO Auto-generated method stub
		super.visitLocalVariable(name, desc, signature, start, end, index);
		}
	}

	@Override
	public void visitEnd() {
		super.visitEnd();
	}

	@Override
	protected void finalize() throws Throwable {
		// TODO Auto-generated method stub
		super.finalize();
	}
	
	
	
}

